/*
 * bmp180.c
 *    Test routine to see how to access the bmp180 via I2C
 *    before implementing the corresponding routines in the bmp180
 *    library from Bosch
 *    Copyright U. Raich Jan. 2017
 *    This program is part of a course on embedded systems held
 *    at the University of Cape Coast, Ghana in 2017
 *    It is released under the GNU Public License 
 *    For details please see https://www.gnu.org/licenses/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdbool.h>
#include <time.h>
#include <linux/types.h>
#include <pigpiod_if2.h>

#define I2C_BUS        1
#define BMP180_ADDRESS 0x77
#define I2C_FLAGS      0

#define BMP180_CTRL_MEAS_REG 0xf4
#define BMP180_OUT_XLSB_REG  0xf8
#define BMP180_OUT_LSB_REG   0xf7
#define BMP180_OUT_MSB_REG   0xf6
#define BMP180_CTRL_MEAS_REG 0xf4
#define BMP180_SOFT_RESET    0xe0
#define BMP180_ID_REG        0xd0
#define BMP180_CALIB_TABLE   0xaa

#define BMP180_SCO(a)        a<<5
#define BMP180_OSS(a)        a<<6

#define BMP180_TEMPERATURE   0x2e
#define BMP180_PRESSURE      0x34
#define BMP180_PRESS_OSS0    0x34
#define BMP180_PRESS_OSS1    0x74
#define BMP180_PRESS_OSS2    0xb4
#define BMP180_PRESS_OSS3    0xf4

#define BMP180_LOW_POWER        0
#define BMP180_STANDARD         1
#define BMP180_HIGH_RES         2
#define BMP180_ULTRA_HIGH_RES   3

typedef	u_int8_t  u8;
typedef	u_int16_t u16;
typedef u_int32_t u32;

typedef int8_t    s8;
typedef	int16_t   s16;
typedef int32_t   s32;

typedef struct  {
  s16 ac1;
  s16 ac2;
  s16 ac3;
  u16 ac4;
  u16 ac5;
  u16 ac6;
  s16 b1;
  s16 b2;
  s16 mb;
  s16 mc;
  s16 md;
} bmp180_calib_t;

double bmp180_calcTemp(u16 rawTemp, bmp180_calib_t *table) {
  double x1,x2, b5, temperature;
  x1= rawTemp - table->ac6;
  x1= x1*(table->ac5)/(double)32768;
  printf("bmp180_calcTemp: x1: %10.4f\n",x1);

  x2=table->mc*(double)2048/(x1 + table->md);
  printf("bmp180_calcTemp: x2: %10.4f\n",x2);

  b5=x1+x2;
  printf("bmp180_calcTemp: b5: %10.4f\n",b5);
  temperature = (b5 + (double) 8)/(double)16;
  return temperature/10.0;
}
double bmp180_calcPressure(u32 rawPressure, u16 rawTemp, u8 oss, bmp180_calib_t *table) {
  s32 pressure,x1,x2,x3,b3,b5,b6;
  u32 b4,b7;
  x1= rawTemp - table->ac6;
  x1= x1*(table->ac5)/32768;
  x2=table->mc*2048/(x1 + table->md);

  b5=x1+x2;
  printf("bmp180_calcPressure: b5: %d\n",b5);
  b6=b5-4000.0;
  printf("bmp180_calcPressure: b6: %d\n",b6);  
  x1 = (table->b2*(b6*b6/4096))/2048;
  printf("bmp180_calcPressure: x1: %d\n",x1);  
  x2 = (table->ac2*b6)/2048;
  printf("bmp180_calcPressure: x2: %d\n",x2);
  x3 = x1+x2;
  printf("bmp180_calcPressure: x3: %d\n",x3);
  b3 = ((((s32)table->ac1*4+x3)<<oss)+2)/4;
  printf("bmp180_calcPressure: b3: %d\n",b3);  
  x1 = table->ac3*b6/8192;
  printf("bmp180_calcPressure: x1: %d\n",x1);  
  x2 = (table->b1*(b6*b6/4096))/65536;
  printf("bmp180_calcPressure: x2: %d\n",x2);  
  x3 = ((x1+x2)+2)/4;
  printf("bmp180_calcPressure: x3: %d\n",x3);  
  b4 = table->ac4*(x3+32768)/32768;
  printf("bmp180_calcPressure: b4: %d\n",b4);    
  b7 = ((u32)rawPressure-b3)*(50000>>oss);
  printf("bmp180_calcPressure: b7: %d\n",b7);  
  if (b7 < 0x80000000)
    pressure = (b7*2)/b4;
  else
    pressure = (b7/b4)*2;
  printf("bmp180_calcPressure: pressure: %d\n",pressure);  
  x1 = (pressure/256)*(pressure/256);
  x1 = x1*3038/65536;
  printf("bmp180_calcPressure: x1: %d\n",x1);  
  x2 = (-7357*pressure)/65536;
  printf("bmp180_calcPressure: x2: %d\n",x2);  
  pressure = pressure + (x1+x2+3791)/16;
  printf("bmp180_calcPressure: pressure: %d\n",pressure);  
  return (double) pressure;
}

u16 swapBytes(u16 input) {
  u16 tmp;
  tmp = input >> 8;
  return tmp | ((input & 0xff) <<8);
}
	  
int bmp180_getCalibTable(int pi, unsigned int handle, bmp180_calib_t *table) {
  int i, retCode;
  u16 value;
  u16 *calibTable = (u16*) table;
  unsigned int reg;
  u8 ac1LSB,ac1MSB,ac1;
  
  reg= BMP180_CALIB_TABLE;
  for (i=0;i<sizeof(bmp180_calib_t)/sizeof(short);i++) {
    value = i2c_read_word_data(pi, handle,reg);
    *calibTable++ = swapBytes(value);
    reg +=sizeof(u16);
  }
  calibTable = (u16*) table;
  for (i=0;i<sizeof(bmp180_calib_t)/sizeof(short);i++) {
    printf("%04x ",*calibTable++);
    printf("\n");
  }
  
  ac1MSB=i2c_read_word_data(pi, handle,BMP180_CALIB_TABLE);
  ac1LSB=i2c_read_word_data(pi, handle,BMP180_CALIB_TABLE+1);
  printf("bmp180_getCalibTable: ac1: %04x\n",(ac1MSB<<8)|ac1LSB);
  return 0;
}

void bmp180_printCalibTable(bmp180_calib_t *table) {
  printf("bmp180Test: calibration table read from bmp180:\n");
  printf("ac1:    %d\n",table->ac1);
  printf("ac2:    %d\n",table->ac2);
  printf("ac3:    %d\n",table->ac3);
  printf("ac4:    %d\n",table->ac4);
  printf("ac5:    %d\n",table->ac5);
  printf("ac6:    %d\n",table->ac6);
  printf("b1:     %d\n",table->b1);
  printf("b2:     %d\n",table->b2);
  printf("mb:     %d\n",table->mb);
  printf("mc:     %d\n",table->mc);
  printf("md:     %d\n",table->md);
}


int main(int argc, char **argv) {
  int retCode,pi,i2cHandle;
  u8 ctrl,oss;
  bmp180_calib_t calibTable;
  u16 rawTemp;
  u8  tempMSB,tempLSB,pressMSB,pressLSB,pressXLSB;
  u32 rawPressure;
  double temperature,pressure;
  
  if ((retCode == pigpio_start(NULL,NULL)) < 0) {
    printf("bmp180Test: Cannot initialize pigpio\n");
    return -1;
  }
  else
    pi=retCode;
  
  printf("bmp180Test: pigpio successfully initialized\n");
  
  if ((retCode = i2c_open(pi,I2C_BUS,BMP180_ADDRESS,I2C_FLAGS)) < 0) {
    printf("bmp180Test: Cannot open I2C bus %d at address %02x\n",
	   I2C_BUS,BMP180_ADDRESS);
    return -1;
  }
  else
    i2cHandle=retCode;
  
  printf("bmp180Test: I2C bus %d successfully opened on address %02x\n",
	 I2C_BUS,BMP180_ADDRESS);

  /*
    read id 
  */
  retCode = i2c_read_byte_data(pi,i2cHandle,BMP180_ID_REG);
  printf("bmp180Test retCode from i2c_read_byte_data of BMP180_ID_REG: %02x\n",
	 retCode);
  /*
    get the calibration table
  */
  retCode = bmp180_getCalibTable(pi,i2cHandle,&calibTable);
  bmp180_printCalibTable(&calibTable);
  /*
    start a conversion
  */
  ctrl= BMP180_SCO(1) | BMP180_OSS(0) | BMP180_TEMPERATURE;
  printf("bmp180Test: control word for temperature meas: %02x\n",ctrl);
  retCode = i2c_write_byte_data(pi,i2cHandle,BMP180_CTRL_MEAS_REG,ctrl);
  if (retCode < 0) {
    printf("bmp180Test: could not write to bmp180 ctrl register\n");
    return -1;
  }
  /*
    read back control register immediately
  */
  retCode = i2c_read_byte_data(pi,i2cHandle,BMP180_CTRL_MEAS_REG);  
  printf("bmp180Test: ctlr register contents: %02x\n",retCode);

  usleep(5000); /* sleep for 5 ms */
  retCode = i2c_read_byte_data(pi,i2cHandle,BMP180_CTRL_MEAS_REG);  
  printf("bmp180Test: ctlr register contents after 5ms wait: %02x\n",retCode);
  /*
    check if conversion is finished 
  */
  if (retCode & BMP180_SCO(1)) {
    printf("bmp180Test: conversion still not finished\n");
    return -1;
  }
  else {
    tempMSB = i2c_read_byte_data(pi,i2cHandle,BMP180_OUT_MSB_REG);
    printf("bmp180Test: temperature: msb: %02x\n",tempMSB);
    tempLSB = i2c_read_byte_data(pi,i2cHandle,BMP180_OUT_LSB_REG);
    printf("bmp180Test: temperature: lsb: %02x\n",tempLSB);
    rawTemp = (tempMSB << 8) | tempLSB;
    printf("bmp180Test: raw temperature: 0x%04x %d\n",rawTemp,rawTemp);
    temperature = bmp180_calcTemp(rawTemp,&calibTable);
    printf("bmp180Test: Temperature: %10.4f °C\n",temperature);
  }
  
  /*
    now read the air pressure with max resolution
  */
  oss = BMP180_ULTRA_HIGH_RES;
  ctrl= BMP180_SCO(1) | BMP180_OSS(oss) | BMP180_PRESSURE;
  printf("bmp180Test: control word for pressure meas: %02x\n",ctrl);
  retCode = i2c_write_byte_data(pi,i2cHandle,BMP180_CTRL_MEAS_REG,ctrl);
  if (retCode < 0) {
    printf("bmp180Test: could not write to bmp180 ctrl register\n");
    return -1;
  }
  usleep(25500); /* sleep for 25.5 ms */
  retCode = i2c_read_byte_data(pi,i2cHandle,BMP180_CTRL_MEAS_REG);  
  printf("bmp180Test: ctlr register contents after 25.5ms wait: %02x\n",retCode);
  /*
    check if conversion is finished 
  */
  if (retCode & BMP180_SCO(1)) {
    printf("bmp180Test: conversion still not finished\n");
    return -1;
  }
  else {
    pressMSB = i2c_read_byte_data(pi,i2cHandle,BMP180_OUT_MSB_REG);
    printf("bmp180Test: pressure: msb: %02x\n",pressMSB);
    pressLSB = i2c_read_byte_data(pi,i2cHandle,BMP180_OUT_LSB_REG);
    printf("bmp180Test: pressure: lsb: %02x\n",pressLSB);
    if (oss==BMP180_ULTRA_HIGH_RES) {
      pressXLSB = i2c_read_byte_data(pi,i2cHandle,BMP180_OUT_XLSB_REG);
      printf("bmp180Test: pressure: lsb: %02x\n",pressXLSB);      
    }
    printf("bmp180Test: 8-oss : %d\n",8-oss); 
    rawPressure = ((pressMSB << 16) | (pressLSB << 8) | pressXLSB) >> (8-oss);
    printf("bmp180Test: raw pressure: 0x%05x %d\n",rawPressure,rawPressure);

  }
  pressure = bmp180_calcPressure(rawPressure, rawTemp, oss, &calibTable);
  printf("bmp180Test: Pressure: %10.4f hPa\n",pressure/100.);
  
  i2c_close(pi,i2cHandle);
  pigpio_stop(pi);  
}
