/*
 * dht11.c: makes a measurement, saving all the bits into an array
 * for later evaluation
 *
 * This is part of a simple meteorological station measuring 
 * temperature, humidity and air pressure every ~ 10s
 * The program will be demonstrated during the UCC Openday on Oct. 18 2017
 * copyright U. Raich Dec 2017
 */

#include <stdio.h>
#include <wiringPi.h>
#include <dht11.h>

#define GPIO_PIN 0

static int protoData[1000]; /* I keep this as global data to give both routines:
                              measure and evaluate, access to it */
static bool debug=false;
static bool writeProtData=false;

short temperature,humidity;

/*
 * switches the debug flag on or off
 */
void dht11SetDebug(bool active) {
  debug = active;
}

void dht11WriteProtocolData(bool active){
  writeProtData = active;
}

/*
 * measure starts a measurement and reads the datastream into the protoData
 * array. There is no error checking here. This is left to the 
 * evaluate function
 */

void dht11Measure() {
  int *dataPtr;
  int i;
  
  wiringPiSetup();
  pinMode(GPIO_PIN,OUTPUT);
  
  if (writeProtData)
    printf("# Raspberry Pi read DHT11 protocol\n");
/* start the dht11 by pulling the data line down for 18 ms 
   and then high for 18 us
*/
  dataPtr = protoData;
  digitalWrite(GPIO_PIN,0);
  delay(18);
  digitalWrite(GPIO_PIN,1);
  delayMicroseconds(40);

  /* switch the GPIO pin to input and read the data every 5 us */
  pinMode(GPIO_PIN,INPUT);
  for (i=0;i<800;i++){
    *dataPtr++ = digitalRead(GPIO_PIN);
    delayMicroseconds(5);
  }
  if (debug) {
    if (writeProtData)
      printf("Write protocol data is switched on\n");
    else
      printf("Write protocol data is switched off\n");
  }      
  /*
   * if in write protocol data is enabled, print out the stuff
   */
  if (writeProtData) {
    printf("dht11Measure: protocol data:\n");
    dataPtr = protoData;
    for (i=0;i<800;i++)
      printf("%d\n",*dataPtr++);
  }
}


/*
 * readBit reads a single bit from the DHT11 data stream.
 */
int readBit(int **data) {
  int retVal,*dataPtr=*data;
  int counter;
  /* wait for the signal to go high again */
  counter = 0;
  /*
  if(debug)
    printf("readBit: pointer to data: 0x%lx\n",(long)dataPtr);
  */
  while(!(*dataPtr)) {
    /*
    if (debug) {
      printf("readBit: dataPtr: 0x%lx *dataPtr: %d\n",(long)dataPtr,*dataPtr);
      printf("readbit: end of protoData: 0x%lx\n",(long)&protoData[999]);
    }
    */
    dataPtr++;    
    counter++;
    if (counter > DHT11_PROTOCOL_TIMEOUT) 
      return (DHT11_PROTOCOL_ERROR);
  }
    /* 
     now look 3 samples further (30 us) on if the pulse is still high
     in which case we have a 1
     else we have a 0
  */
  if (*(dataPtr+5) == HIGH)
    retVal = HIGH;
  else
    retVal =  LOW;
  counter = 0;
  while(*dataPtr) {  /* wait until the signal goes low and the high again */
    /*
    if (debug) {
      printf("readBit: dataPtr: 0x%lx *dataPtr: %d\n",(long)dataPtr,*dataPtr);
      printf("readbit: end of protoData: 0x%lx\n",(long)&protoData[999]);
    }
    */
    dataPtr++;
    counter++;
    if (counter > DHT11_PROTOCOL_TIMEOUT)
      return (DHT11_PROTOCOL_ERROR);
  }
  counter = 0;
  while(!(*dataPtr)) {/* now we are ready to read the next bit */
    /*
    if (debug) {
      printf("readBit: dataPtr: 0x%lx *dataPtr: %d\n",(long)dataPtr,*dataPtr);
      printf("readbit: end of protoData: 0x%lx\n",(long)&protoData[999]);
    }
    */
    dataPtr++;
    counter++;
    if (counter > DHT11_PROTOCOL_TIMEOUT)
      return (DHT11_PROTOCOL_ERROR);
  }
  *data = dataPtr; 
  return retVal;
}

/*
 * evaluate tries to extract the values for temperature and humidity from 
 * the protoData which was filled by measure
 * It tests if the checksum is correct an returns 
 */
int dht11Evaluate() {
  //  int *writePtr;
  int *readPtr,tmp;
  int finalData[5],nextBit;
  int i,j,sum;
    
  /* Now we have to interpret the pulse train */
  /*
  if (debug) {
    printf("dht11Evaluate: protocol data:\n");
    readPtr = protoData;
    for (i=0;i<800;i++)
      printf("%d\n",*readPtr++);
  }
  */
  readPtr = protoData;

  /* 
   *  the protocol header is 80us low (of which only part is seen because
   *  we call the library ro switch the gpio pin from output to input)
   *  followed by 80 us high
   */ 
  while (!(*readPtr++)) ; /* wait until the pulse is high */
  while (*readPtr++) ;    /* wait until it goes high again */
  /*
  while (!(*readPtr))
    printf("%d ",*readPtr++);
  printf("\nbit is high\n");
  while (*readPtr)
    printf("%d ",*readPtr++);
  printf("\nbit is low\n");
  */
  if (debug)    
    printf("dht11Evaluate: after header\n"); 
  for (i=0; i<5;i++) {
    tmp = 0;
    for (j=0;j<8;j++) {
      tmp = tmp << 1;
      /*
      if (debug)
	printf("dht11Evaluate: data pointer before call: %lx\n",(long)readPtr);
      */
      if ((nextBit=readBit(&readPtr)) < 0 ) {
	  printf("Protocol error!\n");
	  return (nextBit);
	}
	else
	  tmp |= nextBit;
      /*
      if (debug)
	printf("dht11Evaluate: data pointer after call: 0x%lx\n",(long)readPtr);
      */
    }
    finalData[i]=tmp;
    /*
    if (debug)
      printf("dht11Evaluate: finalData[%d]: 0x%02x\n",i,tmp);
    */
  }
  sum = 0;
  for (i=0;i<4;i++) {
    if (debug) 
      printf("Data extracted: %x\n",finalData[i]);
    sum += finalData[i];
  }
  temperature = finalData[2];
  humidity    = finalData[0];
  
  if (debug) {
    printf("Checksum calculated: 0x%x, checksum read: 0x%x\n",sum,finalData[4]);
    printf("Temperature: %d.%d\n",finalData[2],finalData[3]);
    printf("Humidity: %d.%d\n",finalData[0],finalData[1]);
  }

  if (sum != finalData[4])
    return (DHT11_BAD_CHECKSUM);
  else
    return(DHT11_SUCCESS);
}

short dht11GetTemp() {
  return temperature;
}

short dht11GetHumidity() {
  return humidity;
}

void dht11PrintError(int err){
  switch (err) {
  case DHT11_SUCCESS:
    printf("dht11 call ended successfully\n");
    break;
  case DHT11_GENERAL_FAILURE:
    printf("General dht11 failure\n");
    break;
   case DHT11_BAD_CHECKSUM:
    printf("dht11 checksum error\n");
    break;
  case DHT11_PROTOCOL_ERROR:
    printf("dht11 checksum error\n");
    break;    
  }
}
