/* @file */

/*! \file  hd44780.c:
 *  \brief 
 *  I2C based PCF8574 access library for HD44780 based LCD on Raspberry Pi
 *  using pigpio for I2C access<br>
 *  Copyright U. Raich Jan. 2017<br>
 *  This program is part of a course on embedded systems held
 *  at the University of Cape Coast, Ghana in 2017<br>
 *  It is released under the GNU Public License 
 *  For details please see https://www.gnu.org/licenses/gpl.html
 *  \author Uli Raich
 */

#include <hd44780.h>
 
static bool hd44780Debug;
static int  pi,i2cHandle=-1;
static unsigned char backlight=HD44780_BL;

static int hd44780WriteCmd(unsigned char cmd);

/*! \brief
 *  switches debug mode on or off
 *  \param onOff true: switches debug mode on<br>
 *               false: switches debug mode off
 */ 
void hd44780SetDebug(bool onOff) {
  hd44780Debug=onOff;
}

int  hd44780BackLight(bool onOff) {
  int retCode;
  unsigned char readBack;
  readBack = i2c_read_byte(pi, i2cHandle);
  if (hd44780Debug)
    printf("hd44780BackLight: readBack first read: 0x%02x\n",readBack);
  if (onOff) {
    readBack |= HD44780_BL;
    backlight = HD44780_BL;
  }
  else {
    readBack &= ~HD44780_BL;
    backlight = 0;
  }
  retCode = i2c_write_byte(pi, i2cHandle, readBack);
  if (retCode < 0) {
    printf("hd44780BackLight: Could not set back light\n");
    return retCode;
  }
  return HD44780_SUCCESS; 
}
/*
 *  send a strobe signal to the EN line
 */
static int hd44780Strobe(unsigned char reg) {
  int retCode;
  unsigned char readBack;
  readBack = i2c_read_byte(pi, i2cHandle);
  if (hd44780Debug)
    printf("hd44780Strobe: readBack first read: 0x%02x\n",readBack);
  readBack = readBack | HD44780_EN | backlight | reg;
  retCode = i2c_write_byte(pi, i2cHandle, readBack);
  if (retCode < 0) {
    printf("hd44780Strobe: Could not write to lcd\n");
    return retCode;
  }

  readBack = i2c_read_byte(pi, i2cHandle);
  if (hd44780Debug)
    printf("hd44780Strobe: readBack second read: 0x%02x\n",readBack);

  readBack = (readBack | backlight | reg) & ~HD44780_EN ;
  retCode = i2c_write_byte(pi, i2cHandle, readBack);
  return HD44780_SUCCESS;
}
/*! \brief Sets the DDRAM address to zero such that the next 
 *          character is written to the top left position if the 
 *          LED screen. If the display is shifted it is set back to
 *          the original position
 */
int hd44780Home() {
  int retCode;
  unsigned char cmd;
    cmd = HD44780_HOME;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: home unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  /*
    wait until homeing is done
  */
  usleep(1520);
  return HD44780_SUCCESS;
}
/*! \brief Sets the DDRAM address to zero such that the next 
 *          character is written to the top left position if the 
 *          LED screen. If the display is shifted it is set back to
 *          the original position
 */
int hd44780Cursor(bool onOff, bool blink) {
  int retCode;
  unsigned char cmd;
  cmd = HD44780_DISPLAY_CURSOR;
  if (onOff)
    cmd |= HD44780_CURSOR_ON;
  if (blink)
    cmd |= HD44780_CURSOR_BLINK;
    
  if (hd44780Debug)
    printf("hd44780Cursor: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Cursor: cursor setting unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(37);
  return HD44780_SUCCESS;
}
/*! \brief Switches the display on or off
 *  \param onOff false: switches the display off<br>
 *               true:  switches it on 
 */
int hd44780Display(bool onOff) {
  int retCode;
  unsigned char cmd;
  cmd = HD44780_DISPLAY_CURSOR;
  if (onOff)
    cmd |= HD44780_DISPLAY_ON;
    
  if (hd44780Debug)
    printf("hd44780Display: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Display: display on/off setting unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(37);
  return HD44780_SUCCESS;
}
/*! \brief Defines it the address counter counts up or down<br>
 *          and sets shifting the whole display on or off
 *   \param incDec 
 *                 false: the address counter is decremented when
 *                        a character is written<br>
 *                 true:  it is incremented
 *   \param shift
 *                 false: display shifts are switched off<br>
 *                 true:  they are switched on  
 */
int hd44780EntryMode(bool incDec, bool shift) {
  int retCode;
  unsigned char cmd;
  cmd = HD44780_ENTRY_MODE;
  if (incDec)
    cmd |= HD44780_CURSOR_INC;
  if (shift)
    cmd |= HD44780_DISPLAY_SHIFT;
    
  if (hd44780Debug)
    printf("hd44780Cursor: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Cursor: cursor setting unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(37);
  return HD44780_SUCCESS;
}
/*! \brief Sets the DDRAM address to the beginning of the second line
 */
int hd44780SecondLine() {
  return (hd44780SetDDRamAddr(HD44780_2ND_LINE));
}
/*! \brief Sets the DDRAM address 
 *  \param addr: the DDRAM address to which the next character will be written
 *               or read from
 */
int hd44780SetDDRamAddr(unsigned char addr) {
  int retCode;
  unsigned char cmd;
    cmd = HD44780_SET_DDRAM | addr;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: home unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(37);
  return HD44780_SUCCESS;
}
/*! \brief Clears the display
 */
int hd44780Clear() {
  int retCode;
  unsigned char cmd;
    cmd = HD44780_CLEAR;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: home unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(37);
  return HD44780_SUCCESS;
}

/*
  write a command to the HD44780
*/

static int hd44780WriteCmd(unsigned char cmd) {
  int retCode;
  unsigned char pcfCmd;
  if (i2cHandle == -1) {
    printf("hd44780Write: I2C library is not initialized!\n");
    printf("Please call hd44780Open first\n");
    return -1;
  }
  if (hd44780Debug)
    printf("hd44780WriteCmd: Command is : 0x%02x\n",cmd);
  /*
    now send high nibble
  */
  pcfCmd = (cmd &0xf0) | backlight | HD44780_WRITE | HD44780_CMD_REG;
  if (hd44780Debug)
    printf("hd44780WriteCmd: Sending high nibble: 0x%02x\n",pcfCmd);
   retCode = i2c_write_byte(pi, i2cHandle, pcfCmd);
   if (retCode < 0) {
     printf("hd44780WriteCmd: Could not write high nibble\n");
     return retCode;
   }
   hd44780Strobe(HD44780_CMD_REG);
  /*
    now send low nibble
  */
   pcfCmd = (cmd << 4) | backlight | HD44780_WRITE | HD44780_CMD_REG;
   if (hd44780Debug)
     printf("hd44780WriteCmd: Sending low nibble: 0x%02x\n",pcfCmd);
   retCode = i2c_write_byte(pi, i2cHandle, pcfCmd);
   if (retCode < 0) {
     printf("hd44780WriteCmd: Could not write low nibble\n");
     return retCode;
   }   
   hd44780Strobe(HD44780_CMD_REG);
   return HD44780_SUCCESS;
}
/*! \brief Writes a character to the display
 *  \param c the character to be written
 */
int hd44780PutC(char c) {
  unsigned char pcfC;
  int retCode;
  
  if (i2cHandle == -1) {
    printf("hd44780PutC: I2C library is not initialized!\n");
    printf("Please call hd44780Open first\n");
    return -1;
  }
  if (hd44780Debug)
    printf("hd44780PutC: Character is : %c\n",c);
  /*
    now send high nibble
  */
  pcfC = (c &0xf0) | backlight | HD44780_WRITE |  HD44780_DATA_REG;
  if (hd44780Debug)
    printf("hd44780PutC: Sending high nibble: 0x%02x\n",pcfC);
   retCode = i2c_write_byte(pi, i2cHandle, pcfC);
   if (retCode < 0) {
     printf("hd44780PutC: Could not write high nibble\n");
     return retCode;
   }
   hd44780Strobe(HD44780_DATA_REG);
  /*
    now send low nibble
  */
   pcfC = (c << 4) | backlight | HD44780_WRITE |  HD44780_DATA_REG;
   if (hd44780Debug)
     printf("hd44780WriteCmd: Sending low nibble: 0x%02x\n",pcfC);
   retCode = i2c_write_byte(pi, i2cHandle, pcfC);
   if (retCode < 0) {
     printf("hd44780WriteCmd: Could not write low nibble\n");
     return retCode;
   }   
   hd44780Strobe(HD44780_DATA_REG);
   usleep(37);
   return HD44780_SUCCESS;
}
/*! \brief Writes a zero terminated character string to the display
 *  Not more than 16 characters are written it no zero is found.
 *  \param text: the zero terminated C character string
 */

int hd44780PutS(char *text) {
  char *txtPtr = text;
  int i,retCode;
  for (i=0;i<HD44780_MAX_STRLEN;i++) {
    if ((retCode=hd44780PutC(*txtPtr++)) < 0)
      return retCode;
    if (*txtPtr == '\0')
      break;
  }
  if (*txtPtr != '\0')
    return -1;
  else
    return HD44780_SUCCESS;
}

static int hd44780Read(unsigned char reg, unsigned char *value) {
  int retCode;
  unsigned char highNibble,lowNibble;
  unsigned char cmd;

  if (i2cHandle == -1) {
    printf("hd44780Write: I2C library is not initialized!\n");
    printf("Please call hd44780Open first\n");
    return -1;
  }
  /*
  read a byte from the hd44780 command register in 4 bit mode
  The high nibble is read first  
  */

  /*
    the upper 4 bits in the command must be set to high in order to make 
    sure the pins are in open drain mode and not pulled down by
    the pcf8574
  */
  cmd =  0xf0 | backlight | HD44780_READ | reg;
  if (hd44780Debug)
    printf("hd44780Read: Setting to read cmd reg: 0x%02x\n",cmd);

  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780Read: Could not set control bits to read cmd reg\n");
    return retCode;
  }
  /* set EN high */
  cmd |= HD44780_EN;
  if (hd44780Debug)
    printf("hd44780Read: Setting enable to high: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780Read: Could not set enable high\n");
    return retCode;
  }

  /* 
     now read high nibble
  */
  retCode = i2c_read_byte(pi, i2cHandle);
  if (retCode < 0) {
    printf("hd44780Read: Could not read high nibble\n");
    return retCode;
  }
  highNibble = retCode;
  if (hd44780Debug)
    printf("hd44780Read: high nibble read: 0x%02x\n",highNibble);
  /* 
     bring EN low and the high again 
  */
  cmd &= ~HD44780_EN;
  if (hd44780Debug)
    printf("hd44780Read: Bring enable down again: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780Read: Could not set enable low\n");
    return retCode;
  }

  cmd |= HD44780_EN;
  if (hd44780Debug)
    printf("hd44780Read: Set enable high for 2nd cycle: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780Read: Could not set enable high\n");
    return retCode;
  }
  /*
    and read the low nibble
  */
  retCode = i2c_read_byte(pi, i2cHandle);  
  if (retCode < 0) {
    printf("hd44780Read: Could not read high nibble\n");
    return retCode;
  }
  lowNibble = retCode;
  if (hd44780Debug)
    printf("hd44780Read: low nibble read: 0x%02x\n",lowNibble);
  /* 
     finally bring enable low again after second cycle
  */
  cmd &= ~HD44780_EN;
  if (hd44780Debug)
    printf("hd44780Read: Bring enable down 2nd cycle: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780Read: Could not set enable low\n");
    return retCode;
  }
  *value = (highNibble &0xf0) | (lowNibble >> 4);
  if (hd44780Debug)
    printf("hd44780Read: value returned: 0x%02x\n",*value);  
  return HD44780_SUCCESS;
}
/*! \brief Reads the busy flag and the DDRAM address
 *  \param *busy:    a pointer to a boolean into which the busy flag will be written
 *  \param *address: a pointer to an unsigned char into which the DDRAM address 
 *                   will be written
 */
int hd44780ReadBusy(bool *busy, unsigned char *address) {
  int retCode;
  unsigned char value;
  retCode =  hd44780Read(HD44780_CMD_REG, &value);
  if (hd44780Debug) {
    printf("hd44780ReadBusy: value: 0x%02x\n",value);  
  }
  if (retCode < 0)
    return retCode;
  *busy = (value & 0x80) ? true : false;
  if (hd44780Debug) {
    if (*busy) 
      printf("hd44780ReadBusy: busy: true\n");  
    else
      printf("hd44780ReadBusy: busy: false\n");
  }
  *address = value & 0x7f;
  if (hd44780Debug)
    printf("hd44780ReadBusy: address: 0x%02x\n",*address);
  return HD44780_SUCCESS;  
}

int hd44780GetC(char *character) {
  int retCode;
  unsigned char value;
  retCode =  hd44780Read(HD44780_DATA_REG, &value);
  if (retCode < 0)
    return retCode;
  *character = value; 
  if (hd44780Debug)
    printf("hd44780GetC: character: 0x%02x %c\n",*character,*character);
  usleep(37);
  return HD44780_SUCCESS;  
}

/* initialize the pigpio library */
/* the display is on I2C address 0x3f */
 
/*! \brief
 *  hd44770Open: opens the I2C connexction to the pcf8574 
 *
 */

int hd44780Open() {
  int i,retCode = 0;
  unsigned char cmd, addr;
  bool busy;
  
  /* 
     initialize the pigpio library 
  */
  if ((retCode = pigpio_start(NULL,NULL)) < 0) {
    printf("hd4470Test: Cannot initialize pigpio\n");
    return retCode;
  }
  else {
    pi=retCode;
    if (hd44780Debug) {
      printf("hd44780Open: pigpio library successfully started\n");
      printf("pi = %d\n",pi);
    }
  }
  
  /* now open the I2C driver and setup the hd44780 I2C address */
  if ((retCode = i2c_open(pi,I2C_BUS,HD44780_ADDRESS,I2C_FLAGS)) < 0) {
    printf("hd44780Test: Cannot open I2C bus %d at address %02x\n",
	   I2C_BUS,HD44780_ADDRESS);
    if ((pigifError_t)retCode == pigif_bad_send)
      printf("Did you start the pigpiod daemon?\n");
    else
      printf("return code was : %d\n",retCode);
    return retCode;
  }
  else {
    i2cHandle=retCode;
    if (hd44780Debug) {
      printf("hd44780Open: I2C bus %d successfully opened on address %02x\n",
	 I2C_BUS,HD44780_ADDRESS);
      printf("i2cHandle = %d\n",i2cHandle);
    }
  }
  
  /*
    Now we follow the instructions for initialization of the hd44780 
    by the MPU
  */
  cmd = 0x30;
  if (hd44780Debug)
    printf("hd44780Test writing 3 x 0x%02x\n",cmd);

  for (i=0;i<5;i++) {
    retCode = i2c_write_byte(pi,i2cHandle,cmd);
    if (retCode < 0) {
      printf("hd44780Test: Could not reset the display\n");
      return retCode;
    }
    hd44780Strobe(HD44780_CMD_REG);
    usleep(4200);      /* wait for 4.2 ms */
  }  
  /*
    now initialze the hd44780 and set it to 4 bit mode
  */
  cmd = HD44780_FUNCTION | HD44780_4_BIT_MODE;
  if (hd44780Debug)
    printf("hd44780Open: hd4478 sending command: 0x%02x\n",cmd);
  
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780Open: Could not write initialization code to hd44780\n");
    return retCode;
  }
  hd44780Strobe(HD44780_CMD_REG);
  /*
    set to 4 bit mode and 5x8 pixel characters
  */

  cmd = HD44780_FUNCTION | HD44780_4_BIT_MODE | HD44780_5x8_FONT | HD44780_2_LINE;
  if (hd44780Debug)
    printf("hd44780Open: sending command: 0x%02x\n",cmd);

  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Open: setting function unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  hd44780Display(HD44780_OFF);
  hd44780Clear();
  hd44780Home();
  /*
    initialization completed
  */
  retCode = hd44780ReadBusy(&busy,&addr);
  if (busy)
    printf("hd44780Open: still busy\n");
  if (hd44780Debug)
    printf("hd44780Open: DDRAM address readback: 0x%02x\n",addr);
  if (addr != 0)
    printf("hd44780Open: DDRAM address readback wrong: 0x%02x\n",addr);

  /*
    switch display back on again
  */
  hd44780Display(HD44780_ON);

  return HD44780_SUCCESS;
}
  
/*
 close the I2C connection
*/
void  hd44780Close() {
  i2c_close(pi,i2cHandle);
  pigpio_stop(pi);
}

/*
  strobe the enable line high and then low again
*/


