/* hd44780Test.c Test program for the Hitachi LCD display driver
 *    Copyright U. Raich Jan. 2017
 *    This program is part of a course on embedded systems held
 *    at the University of Cape Coast, Ghana in 2017
 *    It is released under the GNU Public License 
 *    For details please see https://www.gnu.org/licenses/gpl.html
 */
#include <stdio.h>
#include <pigpiod_if2.h>
#include <hd44780.h>
#include <unistd.h>
#include <stdbool.h>

#define HD44780_SUCCESS 0

bool hd44780Debug=true;
int pi,i2cHandle;

/*
  strobe the enable line high and then low again
*/
int hd44780Strobe(unsigned char reg) {
  int retCode;
  unsigned char readBack;
  readBack = i2c_read_byte(pi, i2cHandle);
  if (hd44780Debug)
    printf("hd44780Strobe: readBack first read: 0x%02x\n",readBack);
  readBack = readBack | HD44780_EN | HD44780_BL | reg;
  retCode = i2c_write_byte(pi, i2cHandle, readBack);
  if (retCode < 0) {
    printf("hd44780Strobe: Could not write to lcd\n");
    return retCode;
  }

  readBack = i2c_read_byte(pi, i2cHandle);
  if (hd44780Debug)
    printf("hd44780Strobe: readBack second read: 0x%02x\n",readBack);

  readBack = (readBack | HD44780_BL | reg) & ~HD44780_EN ;
  retCode = i2c_write_byte(pi, i2cHandle, readBack);
  return HD44780_SUCCESS;
}

int hd44780ReadBusy(bool *busy, unsigned char *addressCnt) {
  int retCode;
  unsigned char highNibble,lowNibble;
  unsigned char cmd;

  if (i2cHandle == -1) {
    printf("hd44780Write: I2C library is not initialized!\n");
    printf("Please call hd44780Open first\n");
    return -1;
  }
  /*
  read a byte from the hd44780 command register in 4 bit mode
  The high nibble is read first  
  */

  cmd =  0xf0 | HD44780_BL | HD44780_READ | HD44780_DATA_REG;
  if (hd44780Debug)
    printf("hd44780ReadBusy: Setting to read cmd reg: 0x%02x\n",cmd);

  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780ReadBusy: Could not set control bits to read cmd reg\n");
    return retCode;
  }
  /* set EN high */
  cmd |= HD44780_EN;
  if (hd44780Debug)
    printf("hd44780ReadBusy: Setting enable to high: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780ReadBusy: Could not set enable high\n");
    return retCode;
  }
  usleep(100);
  /* 
     now read high nibble
  */
  retCode = i2c_read_byte(pi, i2cHandle);
  if (retCode < 0) {
    printf("hd44780ReadCmd: Could not read high nibble\n");
    return retCode;
  }
  highNibble = retCode;
  if (hd44780Debug)
    printf("hd44780ReadBusy: high nibble read: 0x%02x\n",highNibble);
  /* 
     bring EN low and the high again 
  */
  cmd &= ~HD44780_EN;
  if (hd44780Debug)
    printf("hd44780ReadBusy: Bring enable down again: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780ReadBusy: Could not set enable low\n");
    return retCode;
  }
  retCode = i2c_read_byte(pi, i2cHandle);
  if (hd44780Debug)
    printf("hd44780ReadBusy: low nibble read after en low: 0x%02x\n",retCode);
 
  usleep(100);
  cmd |= HD44780_EN;
  if (hd44780Debug)
    printf("hd44780ReadBusy: Set enable high for 2nd cycle: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780ReadBusy: Could not set enable high\n");
    return retCode;
  }
  usleep(100);
  /*
    and read the low nibble
  */
  retCode = i2c_read_byte(pi, i2cHandle);  
  if (retCode < 0) {
    printf("hd44780ReadCmd: Could not read high nibble\n");
    return retCode;
  }
  lowNibble = retCode;
  if (hd44780Debug)
    printf("hd44780ReadBusy: low nibble read: 0x%02x\n",lowNibble);
  /* 
     finally bring enable low again after second cycle
  */
  cmd &= ~HD44780_EN;
  if (hd44780Debug)
    printf("hd44780ReadBusy: Bring enable down 2nd cycle: 0x%02x\n",cmd);
  retCode = i2c_write_byte(pi, i2cHandle, cmd);
  if (retCode < 0) {
    printf("hd44780ReadBusy: Could not set enable low\n");
    return retCode;
  }  
  return HD44780_SUCCESS;
}

int hd44780WriteCmd(unsigned char cmd) {
  int retCode;
  unsigned char pcfCmd;
  if (i2cHandle == -1) {
    printf("hd44780Write: I2C library is not initialized!\n");
    printf("Please call hd44780Open first\n");
    return -1;
  }
  if (hd44780Debug)
    printf("hd44780WriteCmd: Command is : 0x%02x\n",cmd);
  /*
    now send high nibble
  */
  pcfCmd = (cmd &0xf0) | HD44780_BL | HD44780_WRITE | HD44780_CMD_REG;
  if (hd44780Debug)
    printf("hd44780WriteCmd: Sending high nibble: 0x%02x\n",pcfCmd);
   retCode = i2c_write_byte(pi, i2cHandle, pcfCmd);
   if (retCode < 0) {
     printf("hd44780WriteCmd: Could not write high nibble\n");
     return retCode;
   }
   hd44780Strobe(HD44780_CMD_REG);
  /*
    now send low nibble
  */
   pcfCmd = (cmd << 4) | HD44780_BL | HD44780_WRITE | HD44780_CMD_REG;
   if (hd44780Debug)
     printf("hd44780WriteCmd: Sending low nibble: 0x%02x\n",pcfCmd);
   retCode = i2c_write_byte(pi, i2cHandle, pcfCmd);
   if (retCode < 0) {
     printf("hd44780WriteCmd: Could not write low nibble\n");
     return retCode;
   }   
   hd44780Strobe(HD44780_CMD_REG);
   return HD44780_SUCCESS;
}

int hd44780PutC(char c) {
  unsigned char pcfC;
  int retCode;
  
  if (i2cHandle == -1) {
    printf("hd44780PutC: I2C library is not initialized!\n");
    printf("Please call hd44780Open first\n");
    return -1;
  }
  if (hd44780Debug)
    printf("hd44780PutC: Character is : %c\n",c);
  /*
    now send high nibble
  */
  pcfC = (c &0xf0) | HD44780_BL | HD44780_WRITE |  HD44780_DATA_REG;
  if (hd44780Debug)
    printf("hd44780PutC: Sending high nibble: 0x%02x\n",pcfC);
   retCode = i2c_write_byte(pi, i2cHandle, pcfC);
   if (retCode < 0) {
     printf("hd44780PutC: Could not write high nibble\n");
     return retCode;
   }
   hd44780Strobe(HD44780_DATA_REG);
  /*
    now send low nibble
  */
   pcfC = (c << 4) | HD44780_BL | HD44780_WRITE |  HD44780_DATA_REG;
   if (hd44780Debug)
     printf("hd44780WriteCmd: Sending low nibble: 0x%02x\n",pcfC);
   retCode = i2c_write_byte(pi, i2cHandle, pcfC);
   if (retCode < 0) {
     printf("hd44780WriteCmd: Could not write low nibble\n");
     return retCode;
   }   
   hd44780Strobe(HD44780_DATA_REG);
   return HD44780_SUCCESS;
}

int main(int c, char **argv) {
  int retCode=0,i;
  unsigned char cmd, addressCnt;  
  char *hello="Hello World !",*helloPtr;
  bool busy;
  
  if ((retCode = pigpio_start(NULL,NULL)) < 0) {
    printf("hd4470Test: Cannot initialize pigpio\n");
    return -1;
  }
  else
    pi=retCode;
  
  printf("hd44780Test: pigpio successfully initialized\n");

  if ((retCode = i2c_open(pi,I2C_BUS,HD44780_ADDRESS,I2C_FLAGS)) < 0) {
    printf("hd44780Test: Cannot open I2C bus %d at address %02x\n",
	   I2C_BUS,HD44780_ADDRESS);
    if ((pigifError_t)retCode == pigif_bad_send) {
      printf("Did you start the pigpiod daemon?\n");
      return -1;
    }
    else
      printf("return code was : %d\n",retCode);
    switch (retCode) {
    case PI_BAD_I2C_BUS:
      printf("bad i2c bus\n");
      break;
    case PI_BAD_I2C_ADDR:
      printf("bad i2c address\n");
      break;
    case PI_BAD_FLAGS:
      printf("bad i2c flags\n");
      break;
    case PI_NO_HANDLE:
      printf("bad i2c handle missing\n");
      break;
    case PI_I2C_OPEN_FAILED:
      printf("bad i2c flags\n");
      break;
    default:
      printf("unknown error\n");      
    }
    return retCode;
  }
  else
    i2cHandle=retCode;
  
  printf("hd44780Test: I2C bus %d successfully opened on address %02x\n",
	 I2C_BUS,HD44780_ADDRESS);

  /*
    this is needed to make sure the chip is correctly initialized
  */
  cmd = 0x30;
  if (hd44780Debug)
    printf("hd44780Test writing 3 x 0x%02x\n",cmd);
  for (i=0;i<3;i++) {
    retCode = i2c_write_byte(pi,i2cHandle,cmd);
    if (retCode < 0) {
      printf("hd44780Test: Could not reset the display\n");
      goto finish;
    }
    hd44780Strobe(HD44780_CMD_REG);
    usleep(1000);
  }
  
  
  cmd = HD44780_FUNCTION | HD44780_4_BIT_MODE;
  if (hd44780Debug)
    printf("hd44780Test sending command: 0x%02x\n",cmd);

  retCode = i2c_write_byte(pi,i2cHandle,cmd);
  if (retCode < 0) {
    printf("hd44780Test: Could not reset the display\n");
    goto finish;
  }

  hd44780Strobe(HD44780_CMD_REG);

  usleep(5000);

  cmd = HD44780_FUNCTION | HD44780_4_BIT_MODE | HD44780_5x8_FONT | HD44780_2_LINE;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);

  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: setting function unsuccessful: error: %d\n",retCode);
    return retCode;
  }

  usleep(5000);

  cmd = HD44780_HOME;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: home unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(5000);
 
  cmd = HD44780_CLEAR;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);
 
  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: home unsuccessful: error: %d\n",retCode);
    return retCode;
  }
  usleep(5000);

  cmd = HD44780_DISPLAY_CURSOR | HD44780_DISPLAY_ON | HD44780_CURSOR_ON | HD44780_CURSOR_BLINK; 
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);

  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: setting cursor and display unsuccessful: error: %d\n",retCode);
    return retCode;
  }

  /*
    send Hello World !
  */

  helloPtr=hello;
  /*
  while (helloPtr != '\0') {
  */
  for (i=0;i<20;i++) {
    hd44780PutC(*helloPtr++);
    if (*helloPtr == '\0')
      break;
  }

  cmd = HD44780_SET_DDRAM | HD44780_2ND_LINE;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);

  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: switching to second line unsuccessful: error: %d\n",retCode);
    return retCode;
  }

  helloPtr=hello;
  for (i=0;i<20;i++) {
    hd44780PutC(*helloPtr++);
    if (*helloPtr == '\0')
      break;
  }
  
  cmd = HD44780_SET_DDRAM | HD44780_2ND_LINE;
  if (hd44780Debug)
    printf("hd44780Test: sending command: 0x%02x\n",cmd);

  if ((retCode=hd44780WriteCmd(cmd))< 0) {
    printf("hd44780Test: switching to second line unsuccessful: error: %d\n",retCode);
    return retCode;
  }

  retCode = hd44780ReadBusy(&busy, &addressCnt);
  if (retCode < 0)
    printf("hd44780Test: error after hd44780ReadBusy: %d\n",retCode);
finish:
    i2c_close(pi,i2cHandle);
    pigpio_stop(pi);
    return 0;
}
