/* meteo.c Program implementing a simple meteorological station 
 *    
 * This is part of a simple meteorological station measuring 
 * temperature, humidity and air pressure every ~ 10s
 * The program will be demonstrated during the UCC Openday on Oct. 18 2017
 * copyright U. Raich Oct 2017
 */

#include <hd44780.h>
#include <dht11.h>
#include <bmp180.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define MAX_TRY 10

#define Pa2mbar ((double)1013.15 /(double)101325)
extern s8 I2C_routine(void);
extern struct bmp180_t bmp180;

int main(int c, char **argv) {
  FILE *logFile;
  int i,retCode=-1;
  short temperature,humidity;
  char tempString[3],humiString[3];
  u16 rawTemp;
  u32 rawPress;
  double temperatureFloat;
  u32 pressure;
  
  char tempHumiString[] = "Temp:   Humi:  %";
  char pressString[]    = "Press:       hPa";
  char bmp180PressString[7];
  bool successfulTry;
  BMP180_RETURN_FUNCTION_TYPE bmp180RetCode;
  
  /*
    set debugging on
  */
  hd44780SetDebug(false);
  /*
    setup the bmp180
  */
  I2C_routine();
  printf("bmp180: bmp180 structure successfully initialized\n");
  if((bmp180RetCode = bmp180_init(&bmp180)) < 0)
    {
      printf("bmp180: Could not initialize bmp180 chip\n");
      return -1;
    }
  /*
    set to ultra high resolution for pressure measurement and
    average over 3 measurements
  */
  bmp180.oversamp_setting=BMP180_OVERSAMP_SETTING_U8X;
  bmp180.sw_oversamp=BMP180_SW_OVERSAMP_U8X;

  /*
    here we go into an endless loop reading the meteorological data
    and writing them to the display
  */
  successfulTry = false;
  for (;;) {
    for (i=0;i<MAX_TRY;i++) {
      dht11Measure();
      if (dht11Evaluate() < 0)
	sleep(2);
      else {
	temperature = dht11GetTemp();
	humidity    = dht11GetHumidity();
	successfulTry = true;
	break;
      }	
    }
    if (!successfulTry) {
      printf("meteo: %d unsuccessful tries are too much. Giving up...\n",
	     MAX_TRY);
      exit(-1);
    }
    printf("meteo: from dht11: temperature: %d, humidity: %d\n",temperature, humidity);

    /*-------------------------------------------------------------*/  
    /*
      here for the BMP180
      first read the temperature, then the air pressure
    */

    rawTemp = bmp180_get_uncomp_temperature();
    printf("bmp180: raw temperature: %d\n",rawTemp);
    temperatureFloat = (double)bmp180_get_temperature(rawTemp)/10.0;
    printf("bmp180: temperature: %5.1f °C\n",temperatureFloat);
    /*
      now for barometric pressure
    */
    rawPress = bmp180_get_uncomp_pressure(); 
    pressure = bmp180_get_pressure(rawPress);
    printf("bmp180: raw pressure: %d\n",rawPress);
    /*
      101325 Pa <=> 1.01325 bar <=>760 Torr
    */
    printf("bmp180: pressure: %7.2f hPa\n",(double)pressure/100.0);
    sprintf(bmp180PressString,"%7.2f",(double)pressure/100.0);
    for (i=0;i<7;i++)
      pressString[i+6]=bmp180PressString[i];
      
	    
    /*-------------------------------------------------------------*/  
    /*
      save the results to a log file
    */
    if ((logFile=fopen("meteo.dat","a")) == NULL)
      printf("Could not open log file, skipping...\n");
    
    fprintf(logFile,"temperature: %5.2f, humidity: %d%% air pressure %7.2f hPa\n",
	    temperatureFloat,humidity,(double)pressure/100.0);
    if (logFile != NULL)
      fclose(logFile);
    sprintf(tempString,"%02d",(int)round(temperatureFloat));
    sprintf(humiString,"%02d",humidity);
    tempHumiString[5] =tempString[0];
    tempHumiString[6] =tempString[1];
    tempHumiString[13]=humiString[0];
    tempHumiString[14]=humiString[1];
    printf("%s\n",tempHumiString);

    if ((retCode = hd44780Open()) < 0) {
      printf("meteo.c: Cannot initialize pigpio\n");
      return -1;
    }
  

    if (hd44780Home() < 0)
      printf("meteo.c: Could not \"home\" the display\n");
    
    if (hd44780PutS(tempHumiString) < 0)
      printf("meteo.c: Temperature and Humidity failed\n");

    hd44780SecondLine();
    
    if (hd44780PutS(pressString) < 0)
      printf("meteo.c: Air pressurefailed!\n");
    hd44780Close();

    sleep(60);
  }
  
  return HD44780_SUCCESS;
}
  
  
